from .botloggererror import BotLoggerError


class WrongLogLevelError(BotLoggerError):
    """
    Wrong log level exception.

    ...
    Attributes
    ---
    level: str
        inputted log level that does not exist
    """

    def __init__(self, wrong_level: str) -> None:
        """
        Keyword arguments:
        ---
        wrong_level: str, required
            inputted log level that does not exist
        """
        self.level = wrong_level




import os
import json

from .storage import Storage, StorageException
from ..utils.datafiltering import DataFiltering


class LocalJSONStorage(Storage):
    '''
    A class used to work with local storage

    ...
    Attributes
    ---
    address: str
        adress of your mongodb server
    port: int
        port of your mongodb server
    database: str
        name of your database with collections
    collection : str
        collection name from the database
    columns : list
        list of columns to get values from
    doc : dict
        document to insert to the database
    docs : list<dict>
        listr of documents to insert to the database
    value : str
        value from the column
    id_column : str
        name of the id column
    id : str
        id value itself

    Methods
    ---
    get_data(collection, columns)
        Getting list of documents from the certain collection
    get_data_by_column(collection, by, value, columns)
        Getting list of documents from the certain collection by requested column: value
    insert_one_doc(collection, doc)
        Inserts document to the collection
    insert_many_docs(collection, docs)
        Inserts list of documents to the collection
    remove_one_doc_by_column(collection, column, value)
        Removes first found document with specified value from the specified column
    remove_many_docs_by_column(collection, column, value)
        Removes all documents with specified value from the specified column
    update_one_doc(collection, id_column, id, doc)
        Updates first found doc in certain collection by chosen column by dictionary from args
    update_many_docs(collection, id_column, id, doc)
        Updates all documents in certain collection by chosen column by dictionary from args
    '''
    def __init__(self, storage_folder: str) -> None:
        self.__folder = storage_folder + '/'
        self.__init_storage_folder(self.__folder)

    def __init_storage_folder(self, folder: str):
        if not os.path.exists(folder):
            os.makedirs(folder)

    def __write_data_to_json(self, file_name: str, data: list) -> None:
        with open(file_name, 'w') as data_file:
            json.dump(data, data_file)
            data_file.close()

    #---------------------------------------------------------------------------
    # Getting methods
    #---------------------------------------------------------------------------

    def get_data(self, collection: str, columns: list = [], doc: dict = {}, count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        columns : list, optional (default [])
            list of columns to get values from
        '''
        try:
            with open(self.__folder + collection + '.json', 'r') as data_file:
                data = json.load(data_file)
                data_file.close()

            if doc:
                data = filter(lambda x: doc.items() <= x.items(), data)
                data = list(data)

            if columns:
                data = DataFiltering.dict_list_slice(dicts=data,
                                                    columns=columns)
            if count:
                data = data[:count]

            return data
        except:
            raise StorageException("Failed to get data from storage")

    def get_data_by_column(self, collection: str, by: str, value: str, columns: list = [], count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection by requested column: value

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        columns : list, optional (default [])
            list of columns to get values from
        by: str, required
            column name to search by
        value: str, required
            value from the column
        '''
        column_dict = {
            by: value
        }

        find_res = self.get_data(collection=collection,
                                columns=columns,
                                doc=column_dict,
                                count=count)

        return find_res

    #---------------------------------------------------------------------------
    # Insertion methods
    #---------------------------------------------------------------------------

    def insert_one_doc(self, collection: str, doc: dict) -> None:
        '''
        Inserts document to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            document to insert to the database
        '''
        try:
            try:
                data = self.get_data(collection=collection)
            except:
                data = []

            data.append(doc)

            self.__write_data_to_json(self.__folder + collection + '.json', data)

        except:
            raise StorageException('Failed to insert a doc to collection')

    def insert_many_docs(self, collection: str, docs: list) -> None:
        '''
        Inserts list of documents to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        docs : list<dict>, required
            listr of documents to insert to the database
        '''
        try:
            try:
                data = self.get_data(collection=collection)
            except:
                data = []

            data += docs

            self.__write_data_to_json(self.__folder + collection + '.json', data)

        except:
            raise StorageException('Failed to insert a doc to collection')

    #---------------------------------------------------------------------------
    # Removal methods
    #---------------------------------------------------------------------------

    def remove_one_doc_by_column(self, collection: str, column: str, value: str) -> None:
        '''
        Removes first found document with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        self.remove_one_doc_by_dict(collection=collection,
                                    doc={column: value})

    def remove_one_doc_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes first found document containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        try:
            count = 0
            data = self.get_data(collection=collection)
            for dic in data:
                try:
                    if doc.items() <= dic.items():
                        count += 1
                        data.remove(dic)
                        break
                except:
                    continue

            self.__write_data_to_json(self.__folder + collection + '.json', data)
        except:
            raise StorageException('Failed to remove a doc from collection!')

    def remove_many_docs_by_column(self, collection: str, column: str, value: str):
        '''
        Removes all documents with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        self.remove_many_docs_by_dict(collection=collection,
                                    doc={column: value})

    def remove_many_docs_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes all documents containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        try:
            count = 0
            data = self.get_data(collection=collection)
            for dic in data.copy():
                try:
                    if doc.items() <= dic.items():
                        data.remove(dic)
                        count += 1
                except:
                    continue

            self.__write_data_to_json(self.__folder + collection + '.json', data)
        except:
            raise StorageException('Failed to remove a doc from collection!')

    #---------------------------------------------------------------------------
    # Update methods
    #---------------------------------------------------------------------------

    def update_one_doc(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates first found doc in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for the found doc by id
        '''
        try:
            count = 0
            data = self.get_data(collection=collection)
            for index,_ in enumerate(data):
                try:
                    if data[index][id_column] == id:
                        for key in doc.keys():
                            data[index][key] = doc[key]
                        count += 1
                        break
                except:
                    continue

            if count != 0:
                self.__write_data_to_json(self.__folder + collection + '.json', data)

        except:
            raise StorageException('Failed to update a doc from collection!')


    def update_many_docs(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates all documents in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for every found document by id
        '''
        try:
            count = 0
            data = self.get_data(collection=collection)
            for index,_ in enumerate(data):
                try:
                    if data[index][id_column] == id:
                        for key in doc.keys():
                            data[index][key] = doc[key]
                        count += 1
                except:
                    continue
            if count != 0:
                self.__write_data_to_json(self.__folder + collection + '.json', data)
        except:
            raise StorageException('Failed to update a doc from collection!')



from typing import Callable, List


class Route:
    """
    Class managing single income message.

    ..........
    Attributes
    ----------
    callback: Callable, public
        function to be called when route is served
    states: List[str], public
        states one of which user should have to access serving the route
        empty list means that every state is valid
    roles: List[str], public
        roles one of which user should have to access serving the route
        empty list means that every role is valid
    """
    def __init__(self,
                 callback: Callable,
                 states: List[str],
                 roles: List[str]) -> None:
        """
        Constructor.

        .........
        Arguments
        ---------
        callback: Callable, required
            function to be called when route is served
        states: List[str]
            states one of which user should have to access serving the route
        roles: List[str]
            roles one of which user should have to access serving the route
        """
        self.callback = callback
        self.states = states
        self.roles = roles

    def is_accessible_with(self, state: str, roles: List[str]) -> bool:
        """
        Checks if route is accessible to serve with given state and roles.
        returns True in case user's state is in states list AND
                             user is loginned as one of roles in given list

        .........
        Arguments
        ---------
        state: str, required
            user state to check for access to serving the route
        roles: List[str], required
            user roles to check for access to serving the route
        """
        if self.states and state not in self.states:
            return False

        if self.roles:
            for role in self.roles:
                if role in roles:
                    return True
        else:
            return True

        return False




from .role_managing.roleauth import RoleAuth
from .state_managing.statemanager import StateManager
from .user_meta.usermetastorage import UserMetaStorage
from .locales.localemanager import LocaleManager
from .logs.botlogger import BotLogger
from .storage_managing.storage import Storage
from .storage_managing.localjsonstorage import LocalJSONStorage
from .storage_managing.mongodbstorage import MongoDBStorage

from .routing.router import Router
from .routing.commandroute import CommandRoute
from .routing.messageroute import MessageRoute

from .messaging.messagesender import MessageSender

from .exceptions.telegramboterror import TelegramBotError



import re
from typing import List

import telegram as tg
import telegram.ext as tg_ext

from ..state_managing.statemanager import StateManager
from ..state_managing.exceptions.stateerror import StateError
from ..role_managing.roleauth import RoleAuth
from ..role_managing.exceptions.roleerror import RoleError

from .commandroute import CommandRoute
from .messageroute import MessageRoute


class Router:
    """
    Class managing income messages callbacks

    ..........
    Attributes
    ----------
    message_routes: List[MessageRoute], private
        list of text message routes
    command_routes: List[CommandRoute], private
        list of commands routes (commands start with /)
    tg: telegram.ext.Dispatcher, private
        `python-telegram-bot` class, dispatching all kinds of updates
    state_manager: StateManager, private
        class managing user states
    role_auth: RoleAuth, private
        class managing user roles

    .......
    Methods
    -------
    register_command_route(): public
        registers given command handler
    register_message_route(): public
        registers given message handler
    find_command_routes(): private
        finds all command routes that are accessible with user state and roles and triggered with command
    find_message_routes(): private
        finds all message routes that are accessible with user state and roles and triggered with message
    serve_command_route(): private
        handler for all registered commands, distributes callback functions to received commands
    serve_message_route(): private
        handler for all registered messages, distributes callback functions to received messages
    """
    def __init__(self,
                 tg_dispatcher: tg_ext.Dispatcher,
                 state_manager: StateManager,
                 role_auth: RoleAuth) -> None:
        """
        Constructor.

        .........
        Arguments
        ---------
        tg_dispatcher: telegram.ext.Dispatcher, required
            `python-telegram-bot` class, dispatching all kinds of updates
        state_manager: StateManager, required
            state system
        role_auth: RoleAuth, required
            role/authorization system
        """
        self.__message_routes: List[MessageRoute] = []
        self.__command_routes: List[CommandRoute] = []

        self.__tg = tg_dispatcher
        self.__state_manager = state_manager
        self.__role_auth = role_auth

    def __find_command_routes(self,
                              command: str,
                              state: str,
                              roles: List[str]) -> List[CommandRoute]:
        """
        Finds all command routes that are accessible with user state and roles and triggered with command

        .........
        Arguments
        ---------
        command: str, requited
            trigger command to serve the route
        state: str, required
            user state for access to serving the route
        roles: List[str], required
            user roles for access to serving the route
        """
        found_routes = []

        for command_route in self.__command_routes:
            if command == command_route.command and\
               command_route.is_accessible_with(state, roles):
                found_routes.append(command_route)
        return found_routes

    def __find_message_routes(self,
                              message: str,
                              state: str,
                              roles: List[str]) -> List[MessageRoute]:
        """
        Finds all message routes that are accessible with user state and roles and triggered with message

        .........
        Arguments
        ---------
        message: str, requited
            trigger regex message text to serve the route
        state: str, required
            user state for access to serving the route
        roles: List[str], required
            user roles for access to serving the route
        """
        found_routes = []

        for message_route in self.__message_routes:
            if re.fullmatch(message_route.message, message) and\
               message_route.is_accessible_with(state, roles):
                found_routes.append(message_route)
        return found_routes

    def __serve_command_route(self, update: tg.Update, context: tg_ext.CallbackContext) -> None:
        """
        Handler for all registered commands, distributes callback functions to received commands.
        command is considered to start with '/' and end with end of string or space.
        Calls callback function with **kwargs:
            user_id: int
                chat id that command is received from
            message: str
                full received message text
            username: str
                telegram username of user with user_id
            first_name: str
                telegram first name of user with user_id
            last_name: str
                telegram last name of user with user_id
            roles: List[str]
                roles of user with user_id
            state: Union[str, dict]
                state of user presented by state_name if 'state_with_params' was True,
                otherwise it is presented as dict with keys 'State' and 'State_Params' by default

        .........
        Arguments
        ---------
        update: telegram.Update, required
            `python-telegram-bot` class, representing an incoming update
        context: telegram.ext.CallbackContext, required
            `python-telegram-bot` class. context passed by telegram handler to callback
        """
        message = update.message
        chat = message.chat

        message = message.text
        user_id = chat.id
        username = chat.username
        first_name = chat.first_name
        last_name = chat.last_name

        command = message.split()[0].strip('/')
        try: # try-except for user start initialization when he is not in db
            user_state = self.__state_manager.get_state(user_id)
            user_roles = self.__role_auth.get_user_roles(user_id)
        except (StateError, RoleError):
            user_state = 'free'
            user_roles = ['user']

        found_routes = self.__find_command_routes(command, user_state, user_roles)

        for route in found_routes:
            route.callback(user_id=user_id,
                           message=message,
                           username=username,
                           first_name=first_name,
                           last_name=last_name,
                           roles=user_roles,
                           state=user_state)

    def __serve_message_route(self, update: tg.Update, context: tg_ext.CallbackContext) -> None:
        """
        Handler for all registered messages, distributes callback functions to received messages.
        Calls callback function with **kwargs:
            user_id: int
                chat id that command is received from
            message: str
                full received message text
            username: str
                telegram username of user with user_id
            first_name: str
                telegram first name of user with user_id
            last_name: str
                telegram last name of user with user_id
            roles: List[str]
                roles of user with user_id
            state: Union[str, dict]
                state of user presented by state_name if 'state_with_params' was True,
                otherwise it is presented as dict with keys 'State' and 'State_Params' by default

        .........
        Arguments
        ---------
        update: telegram.Update, required
            `python-telegram-bot` class, representing an incoming update
        context: telegram.ext.CallbackContext, required
            `python-telegram-bot` class. context passed by telegram handler to callback
        """
        message = update.message
        chat = message.chat

        message = message.text
        user_id = chat.id
        username = chat.username
        first_name = chat.first_name
        last_name = chat.last_name

        try: # try-except for user start initialization when he is not in db
            user_state = self.__state_manager.get_state(user_id)
            user_roles = self.__role_auth.get_user_roles(user_id)
        except (StateError, RoleError):
            user_state = 'free'
            user_roles = ['user']

        found_routes = self.__find_message_routes(message, user_state, user_roles)

        for route in found_routes:
            route.callback(user_id=user_id,
                           message=message,
                           username=username,
                           first_name=first_name,
                           last_name=last_name,
                           roles=user_roles,
                           state=user_state)

    def register_command_route(self, route: CommandRoute) -> None:
        """
        Registers given command handler.

        .........
        Arguments
        ---------
        route: CommandRoute, required
            route to register
        """
        self.__command_routes.append(route)
        self.__tg.add_handler(tg_ext.CommandHandler(command=route.command,
                                                    callback=self.__serve_command_route))

    def register_message_route(self, route: MessageRoute) -> None:
        """
        Registers given text message handler.

        .........
        Arguments
        ---------
        route: MessageRoute, required
            route to register
        """
        self.__message_routes.append(route)
        self.__tg.add_handler(tg_ext.MessageHandler(filters=tg_ext.Filters.regex(route.message),
                                                    callback=self.__serve_message_route))




from ..storage_managing.storage import Storage
from .exceptions.usermetaerror import UserMetaError


class UserMetaStorage:
    '''
    Class for users' meta data handling
    '''
    def __init__(self, storage: Storage, users_collection: str = "Users",
                id_column: str = "_id", username_column: str = "Username",
                first_name_column: str = "First_Name", last_name_column: str = "Last_Name",
                locale_column: str = "Locale"):
        '''
        Parameters
        ----------
        storage : Storage
            pointer to the inherited class from Storage (i.e MongoDBStorage)
        users_collection : str, optional
            name of id column from users_collection, by default "Users"
        id_column : str, optional
            name of id column from users_collection, by default "_id"
        locale_column : str, optional
            name of user's locale column from users_collection, by default "Locale"
        '''
        self.__users_collection = users_collection
        self.__id_column = id_column
        self.__username_column = username_column
        self.__first_name_column = first_name_column
        self.__last_name_column = last_name_column
        self.__locale_column = locale_column

        self.__storage = storage

    def get_locale(self, user_id: int) -> str:
        '''
        Gives a locale of the requested user

        Parameters
        ----------
        user_id : int
            User's id from the storage

        Returns
        -------
        str
            locale code name

        Raises
        ------
        UserMetaError
            Raises an exception if something went wrong
        '''
        locales = self.__storage.get_data_by_column(collection=self.__users_collection,
                                                    by=self.__id_column,
                                                    value=user_id)
        if not locales:
            raise UserMetaError("Meta Exception")

        locale = locales[0]

        return locale[self.__locale_column]

    def user_exists(self, user_id: int) -> bool:
        '''
        Checks whether user exists or not

        Parameters
        ----------
        user_id : int
            User's id from the storage

        Returns
        -------
        bool
            "True" if user exists else "False"
        '''
        users = self.__storage.get_data_by_column(self.__users_collection, self.__id_column, user_id)

        return True if users else False

    def user_initialize(self, user_id: int, init_dict: dict):
        '''
        User initialization method. Should be  used to add new user to the storage.

        Parameters
        ----------
        user_id : int
            User's id
        init_dict : dict
            Initialization dictionary
            ... example:
            {
                "Locale": "en_us",
                "Roles": ["user"],
                "State": "free",
                "State_Params": {}
            }
        '''
        if not self.user_exists(user_id):
            init_dict[self.__id_column] = user_id

            self.__storage.insert_one_doc(self.__users_collection, init_dict)

    def user_update(self, user_id: int, username: str = "", first_name: str = "", last_name: str = ""):
        '''
        Method for updating user's data in the storage.

        Parameters
        ----------
        user_id : int
            User's id from the storage
        username : str, optional
            User's username, by default ""
        first_name : str, optional
            User's first name, by default ""
        last_name : str, optional
            User's last name, by default ""

        Raises
        ------
        UserMetaError
            [description]
        '''
        upd_dict = {
            self.__username_column: username,
            self.__first_name_column: first_name,
            self.__last_name_column: last_name
        }
        try:
            self.__storage.update_one_doc(self.__users_collection, self.__id_column, user_id, upd_dict)
        except:
            raise UserMetaError("Failed to update user")

    def get_username(self, user_id: int) -> str:
        '''
        Gives user's username

        Parameters
        ----------
        user_id : int
            User's id from the storage
        Returns
        -------
        str
            username
        '''
        try:
            user_id = int(user_id)

            usernames_found = self.__storage.get_data_by_column(self.__users_collection,
                                                            by=self.__id_column,
                                                            value=user_id,
                                                            columns=[self.__username_column])

            username = usernames_found[0]
            username = username[self.__username_column]
            return username
        except:
            raise UserMetaError("User was not found!")

    def get_first_name(self, user_id: int) -> str:
        '''
        Gives user's first name

        Parameters
        ----------
        user_id : int
            User's id from the storage
        Returns
        -------
        str
            first name
        '''
        try:
            user_id = int(user_id)

            first_name_found = self.__storage.get_data_by_column(self.__users_collection,
                                                                by=self.__id_column,
                                                                value=user_id,
                                                                columns=[self.__first_name_column])

            first_name = first_name_found[0]
            first_name = first_name[self.__first_name_column]
            return first_name
        except:
            raise UserMetaError("User was not found!")

    def get_last_name(self, user_id: int) -> str:
        '''
        Gives user's last name

        Parameters
        ----------
        user_id : int
            User's id from the storage
        Returns
        -------
        str
            last name
        '''
        try:
            user_id = int(user_id)

            last_name_found = self.__storage.get_data_by_column(self.__users_collection,
                                                                by=self.__id_column,
                                                                value=user_id,
                                                                columns=[self.__last_name_column])

            last_name = last_name_found[0]
            last_name = last_name[self.__last_name_column]
            return last_name
        except:
            raise UserMetaError("User was not found!")




from datetime import datetime
from typing import Union


class DataFiltering:
    '''
    Simple auxiliary filtering class
    '''
    @staticmethod
    def dict_list_slice(dicts: list, columns: list) -> list:
        '''
        Slices list of dicts by given columns(keys)

        Parameters
        ----------
        dicts : list
            list of dicts
        columns : list
            column(keys) names

        Returns
        -------
        list
            list of dicts sliced by given keys
        '''
        if columns != []:
            for i, _ in enumerate(dicts):
                dicts[i] = { c:dicts[i][c] for c in columns }
            return dicts
        else:
            return dicts

    @staticmethod
    def cyrillic_to_latin(text: str) -> str:
        '''
        Changes all cyrllic letter to its latin analog

        Parameters
        ----------
        text : str
            text where you want to change cyrillic letters

        Returns
        -------
        str
            modified string
        '''
        sim_dict = {
            "А": "A", "Р": "P", "К": "K",
            "В": "B", "Т": "T",
            "С": "C", "Х": "X",
            "Е": "E", "О": "O",
            "Н": "H", "М": "M"
        }

        for sym in sim_dict.keys():
            text = text.replace(sym, sim_dict[sym])

        return text

    @staticmethod
    def timestamp_to_date(date: Union[str, int]) -> str:
        '''
        transforms timestamp date to Y-m-d H:M:S format

        Parameters
        ----------
        date : Union[str, int]
            timestamp string or int

        Returns
        -------
        str
            formatted date string in Y-m-d H:M:S format
        '''
        ts = int(date)
        date = datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

        return str(date)



from abc import ABC, abstractmethod

from .exceptions.storageexception import StorageException


class Storage(ABC):
    '''
    Base class for storing data
    '''
    #---------------------------------------------------------------------------
    # Getting methods
    #---------------------------------------------------------------------------

    @abstractmethod
    def get_data(self, collection: str, columns: list = [], doc: dict = {}, count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        columns : list, optional (default [])
            list of columns to get values from
        '''
        pass

    @abstractmethod
    def get_data_by_column(self, collection: str, by: str, value: str, columns: list = [], count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection by requested column: value

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        columns : list, optional (default [])
            list of columns to get values from
        by: str, required
            column name to search by
        value: str, required
            value from the column
        '''
        pass

    #---------------------------------------------------------------------------
    # Insertion methods
    #---------------------------------------------------------------------------

    @abstractmethod
    def insert_one_doc(self, collection: str, doc: dict) -> None:
        '''
        Inserts document to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        doc : dict, required
            document to insert to the storage
        '''
        pass

    @abstractmethod
    def insert_many_docs(self, collection: str, docs: list) -> None:
        '''
        Inserts list of documents to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        docs : list<dict>, required
            listr of documents to insert to the storage
        '''
        pass

    #---------------------------------------------------------------------------
    # Removal methods
    #---------------------------------------------------------------------------

    @abstractmethod
    def remove_one_doc_by_column(self, collection: str, column: str, value: str) -> None:
        '''
        Removes first found document with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        pass

    @abstractmethod
    def remove_one_doc_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes first found document containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        pass

    @abstractmethod
    def remove_many_docs_by_column(self, collection: str, column: str, value: str):
        '''
        Removes all documents with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        pass

    @abstractmethod
    def remove_many_docs_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes all documents containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        pass

    #---------------------------------------------------------------------------
    # Update methods
    #---------------------------------------------------------------------------

    @abstractmethod
    def update_one_doc(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates first found doc in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for the found doc by id
        '''
        pass

    @abstractmethod
    def update_many_docs(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates all documents in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for every found document by id
        '''
        pass



import time

from ..storage_managing.storage import Storage
from .exceptions.wrongloglevelerror import WrongLogLevelError


class BotLogger:
    """
    Telegram bot logger writing into DB

    ...
    Attributes
    ---
    levels: dict, static
        available log levels enum
    current_level: str
        current log level: messages with levels higher or equal to current are sent
    collection: str
        DB collection name to send logs to
    storage: Storage
        DB wrapper to send or dump logs
    short_params: bool
        one-string log params flag (params are cut if True)

    Methods
    ---
    cut_params(params)
        cuts params values to a single string
    log_cleaning()
        logs cleaning old logs from DB with info level
    write_log(level, event, params)
        sends log to DB with regard to current log level and flag values
    clean_old(min_age)
        deletes logs older than given age
    dump_last(count)
        dumps given count of last sent logs
    log_error(event, params)
        logs given event with error level
    log_receive_command(id, command)
        logs receiving command with info level
    log_receive_document(id, filename)
        logs receiving document with info level
    log_receive_msg(id, text)
        logs receiving message with info level
    log_send_document(id, filename)
        logs sending document with info level
    log_send_msg(id, text)
        logs sending message with info level
    log_start()
        logs starting bot with info level
    log_stop()
        logs stopping bot with info level
    set_level(level)
        sets current log level to given
    """

    levels = {'DEBUG': 1, 'INFO': 2, 'WARNING': 3, 'ERROR': 4, 'CRITICAL': 5}

    def __init__(self,
                 storage: Storage,
                 collection_name: str) -> None:
        """
        Keyword arguments:
        ---
        storage: Storage, required
            DB storage sevice
        collection_name: str
            name of DB collection for logs
        """
        self.__current_level: str = 'INFO'

        self.__collection: str = collection_name
        self.__storage: Storage = storage

        self.short_params: bool = True

    def __cut_params(self, params: dict) -> dict:
        """
        Cuts params values to a single string. If param value contains at least one "\n",
        replacing it with all content afterwards with " <...>".

        Keyword arguments:
        ---
        params: dict, required params which values need to be cut to a single string
        """
        for param in params:
            params[param] = str(params[param])
            if '\n' in params[param]:
                params[param] = params[param].split('\n')[0] + ' <...>'
        return params

    def __log_cleaning(self) -> None:
        """
        Logs cleaning old logs from DB with info level.
        """
        self.__write_log('INFO', 'Old logs cleaned')


    def __write_log(self, level: str, event: str, params: dict = None) -> None:
        """
        Sends log to DB with regard to current log level and flag values.

        Keyword arguments:
        ---
        level: str, required
            level to send log with (if less than current, does not send log)
        event: str, required
            main log description
        params: dict, optional
            additional params to log
        """
        if self.levels[self.__current_level] > self.levels[level]:
            return

        document = {
                    'time': int(time.time() * 1000),
                    'event': event
                   }
        if params:
            if self.short_params:
                params = self.__cut_params(params)
            document['params'] = params

        self.__storage.insert_one_doc(self.__collection, document)


    def clean_old(self, min_age: int) -> None:
        """
        Deletes from DB logs older than given age

        Keyword arguments:
        ---
        min_age: int, required
            minimal age (in milliseconds) to delete logs from
        """
        max_time = int(time.time() * 1000) - min_age
        self.__storage.remove_many_docs_by_dict(self.__collection, {'time': {'$lte': max_time}})
        self.__log_cleaning()

    def dump_last(self, count: int) -> list:
        """
        Dumps from DB given count of last sent logs

        Keyword arguments:
        ---
        count: int, required
            number of logs to dump
        """
        return self.__storage.get_data(self.__collection)[-count:]

    def log_error(self, event: str, params: dict = None) -> None:
        """
        Logs given event with error level.

        Keyword arguments
        ---
        event: str, requrired
            main log description
        params: dict, optional
            additional params to log
        """
        self.__write_log('ERROR', event, params)

    def log_receive_command(self, id: int, command: str) -> None:
        """
        Logs receiving command with info level.

        Keyword arguments
        ---
        id: int, required
            telegram id of user who sent command
        command: str, required
            received command
        """

        self.__write_log('INFO',
                         'Command received',
                         {'id': id, 'command': command})

    def log_receive_document(self, id: int, filename: str) -> None:
        """
        Logs receiving document with info level.

        Keyword arguments
        ---
        id: int, required
            telegram id of user who sent document
        filename: str, required
            filename of received document
        """
        self.__write_log('INFO',
                         'Document received',
                         {'id': id, 'filename': filename})

    def log_receive_msg(self, id: int, text: str) -> None:
        """
        Logs receiving message with info level.

        Keyword arguments
        ---
        id: int, required
            telegram id of user who sent message
        text: str, required
            received message text
        """
        self.__write_log('INFO',
                         'Message received',
                         {'id': id, 'text': text})

    def log_send_document(self, id: int, filename: str) -> None:
        """
        Logs sending document with info level.

        Keyword arguments
        ---
        id: int, required
            telegram id of receipient
        filename: str, required
            sent document filename
        """
        self.__write_log('INFO',
                         'Document sent',
                         {'id': id, 'filename': filename})

    def log_send_msg(self, id: int, text: str) -> None:
        """
        Logs sending message with info level.

        Keyword arguments
        ---
        id: int, required
            telegram id of receipient
        text: str, required
            sent message text
        """
        self.__write_log('INFO',
                         'Message sent',
                         {'id': id, 'text': text})

    def log_start(self) -> None:
        """
        Logs starting bot with info level.
        """
        self.__write_log('INFO',
                         'Bot started')

    def log_stop(self) -> None:
        """
        Logs stopping bot with info level.
        """
        self.__write_log('INFO',
                         'Bot stopped')

    def set_level(self, level: str) -> None:
        """
        Sets current log level to given. If given log level does not exist, exception is raised

        Keyword arguments
        ---
        level: str, required
            log level that current level must be

        Raisable exceptions
        ---
        WrongLogLevelError
            raised if given log level does not exist
        """
        if level in self.levels:
            self.__current_level = level
        else:
            raise WrongLogLevelError(level)




from typing import Union

from ..storage_managing.storage import Storage
from .exceptions.stateerror import StateError


class StateManager:
    '''
    Class for managing users' state
    '''
    def __init__(self, storage: Storage, users_collection: str = "Users",
                    id_column: str = "_id", status_column: str = "State",
                    params_column: str = "State_Params", free_status: str = "free",
                    with_params: bool = False):
        '''
        Parameters
        ----------
        storage : Storage
            pointer to the inherited class from Storage (i.e MongoDBStorage)
        users_collection : str, optional
            name of id column from users_collection, by default "Users"
        id_column : str, optional
            name of id column from users_collection, by default "_id"
        status_column : str, optional
            name of status column from users_collection, by default "State"
        params_column : str, optional
            name of state parameters column from users_collection, by default "State_Params"
        free_status : str, optional
            value of "free" status, by default "free"
        with_params : bool, optional
            whether to use params or not, by default False
        '''
        self.__users_collection = users_collection
        self.__id_column = id_column
        self.__status_column = status_column
        self.__params_column = params_column

        self.__free_status = free_status

        self.__storage = storage
        self.__with_params = with_params

    def get_state(self, user_id: int) -> str:
        '''
        Gives a state of specified user

        Parameters
        ----------
        user_id : int
            User's id from the storage

        Returns
        -------
        str
            user's status string

        Raises
        ------
        StateError
            Raises an exception if user was not found
        '''
        # if self.__with_params:
        #     columns = [self.__status_column, self.__params_column]
        # else:
        columns = [self.__status_column]

        response = self.__storage.get_data_by_column(collection=self.__users_collection,
                                                    by=self.__id_column,
                                                    value=user_id,
                                                    columns=columns)
        if not response:
            raise StateError("User was not found!")

        state = response[0]

        # if self.__with_params:
        #     return state
        # else:
        return state[self.__status_column]

    def get_state_params(self, user_id: int) -> dict:
        '''
        Gives a state params of specified user

        Parameters
        ----------
        user_id : int
            User's id from the storage

        Returns
        -------
        dict
            user's state_params if with_params = True

        Raises
        ------
        StateError
            Raises an exception if user was not found
        '''
        if not self.__with_params:
            raise StateError("Class was initialized with no params")
        columns = [self.__params_column]

        response = self.__storage.get_data_by_column(collection=self.__users_collection,
                                                    by=self.__id_column,
                                                    value=user_id,
                                                    columns=columns)
        if not response:
            raise StateError("User was not found")

        state = response[0]
        return state[self.__params_column]

    def is_free(self, user_id: int) -> bool:
        '''
        Checks if user's status is set to "free"

        Parameters
        ----------
        user_id : int
            User's id from the storage

        Returns
        -------
        bool
            wheter user's status if "free" or not
        '''
        user_status = self.get_state(user_id)

        if self.__with_params:
            return True if user_status[self.__status_column] == self.__free_status else False
        else:
            return True if user_status == self.__free_status else False

    def set_state(self, user_id: int, status: str, params: dict = {}) -> None:
        '''
        Set user's status (and user's state_params if with_params = True)

        Parameters
        ----------
        user_id : int
            User's id from the storage
        status : str, dict
            user's status string
        params : dict
            user's state_params (if with_params = True)

        Raises
        ------
        StateError
            Raises an exception if something went wrong trying to change user's status (i.e wrong id)
        '''
        try:
            status_doc = {
                self.__status_column: status
            }

            if self.__with_params:
                status_doc[self.__params_column] = params

            self.__storage.update_one_doc(collection=self.__users_collection,
                                        id_column=self.__id_column,
                                        id=user_id,
                                        doc=status_doc)

        except:
            raise StateError(f'Failed to change {user_id}\'s state to {status}')

    def set_free(self, user_id: int) -> None:
        '''
        Sets user's status to "free" (and state_params to {} if with_params = True)

        Parameters
        ----------
        user_id : int
            User's id from the storage
        '''
        if self.__with_params:
            self.set_state(user_id=user_id,
                        status=self.__free_status,
                        params={})
        else:
            self.set_state(user_id=user_id,
                        status=self.__free_status)



from typing import List

import telegram as tg
import telegram.ext as tg_ext

from ..logs.botlogger import BotLogger


class MessageSender:
    """
    Class that sends messages to users.

    ..........
    Attributes
    ----------
    updater: telegram.ext.Updater, private
        `python-telegram-bot` class, providing a frontend to telegram bot
    logger: BotLogger, private
        logger/logs manager

    .......
    Methods
    -------
    send_text(): public
        sends text message to user
    """
    def __init__(self,
                 updater: tg_ext.Updater,
                 logger: BotLogger) -> None:
        """
        Constructor.

        .........
        Arguments
        ---------
        updater: telegram.ext.Updater, required
            `python-telegram-bot` class, providing a frontend to telegram bot
        logger: BotLogger, required
            logger/logs manager
        """
        self.__updater = updater
        self.__logger = logger

    def send_text(self,
                  user_id: int,
                  text: str,
                  reply_keyboard: List[List[str]] = None,
                  reply_keyboard_resize = True,
                  one_time_keyboard = True) -> None:
        """
        Sends text message to user.

        .........
        Arguments
        ---------
        user_id: int, required
            id of user whom to send message
        text: str, required
            text of sending message
        reply_keyboard: List[List[str]], optional (default = None)
            reply keyboard buttons' texts
        reply_keyboard_resize: bool, optional (default = True)
            flag that requests clients to resize keyboard
        one_time_keyboard: bool, optional (default = True)
            flag that removes older keyboard or hides present one after it is used
        """
        if reply_keyboard is None:
            if one_time_keyboard:
                reply_keyboard = tg.ReplyKeyboardRemove()
        else:
            reply_keyboard = tg.ReplyKeyboardMarkup(keyboard=reply_keyboard,
                                                    resize_keyboard=reply_keyboard_resize,
                                                    one_time_keyboard=one_time_keyboard)

        message = self.__updater.bot.send_message(chat_id=user_id,
                                                  text=text,
                                                  reply_markup=reply_keyboard)
        self.__logger.log_send_msg(id=user_id, text=text)




import os

from pymongo import MongoClient

from .storage import Storage, StorageException
from ..utils.datafiltering import DataFiltering


class MongoDBStorage(Storage):
    '''
    A class used to work with MongoDB

    ...
    Attributes
    ---
    address: str
        adress of your mongodb server
    port: int
        port of your mongodb server
    database: str
        name of your database with collections
    collection : str
        collection name from the database
    columns : list
        list of columns to get values from
    doc : dict
        document to insert to the database
    docs : list<dict>
        listr of documents to insert to the database
    value : str
        value from the column
    id_column : str
        name of the id column
    id : str
        id value itself

    Methods
    ---
    get_data(collection, columns)
        Getting list of documents from the certain collection
    get_data_by_column(collection, by, value, columns)
        Getting list of documents from the certain collection by requested column: value
    insert_one_doc(collection, doc)
        Inserts document to the collection
    insert_many_docs(collection, docs)
        Inserts list of documents to the collection
    remove_one_doc_by_column(collection, column, value)
        Removes first found document with specified value from the specified column
    remove_many_docs_by_column(collection, column, value)
        Removes all documents with specified value from the specified column
    update_one_doc(collection, id_column, id, doc)
        Updates first found doc in certain collection by chosen column by dictionary from args
    update_many_docs(collection, id_column, id, doc)
        Updates all documents in certain collection by chosen column by dictionary from args
    '''
    def __init__(self,
                address: str = "localhost",
                port: int = 27017,
                username: str = "",
                password: str = "",
                database: str = "config"):
        '''
        Keyword arguments:
        ---
        address: str, optional (default "localhost")
            adress of your mongodb server
        port: int, optional (default 27017)
            port of your mongodb server
        database: str, optional (default "config")
            name of your database name with collection
        '''
        self.__init_mongo_connection(address=address,
                                    port=port,
                                    username=username,
                                    password=password)
        self.__database = self.__client[database]

    def __init_mongo_connection(self, address: int, port: int, username: str, password: str) -> None:
        try:
            self.__client = MongoClient(address, port,
                                        username=username,
                                        password=password)

        except:
            raise StorageException('Failed to connect to mongo!')

    #---------------------------------------------------------------------------
    # Getting methods
    #---------------------------------------------------------------------------

    def get_data(self, collection: str, columns: list = [], doc: dict = {}, count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        columns : list, optional (default [])
            list of columns to get values from
        '''

        try:
            db_collection = self.__database[collection]

            if count:
                db_data = list(db_collection.find(doc).limit(count))
            else:
                db_data = list(db_collection.find(doc))

            db_data_filtered = DataFiltering.dict_list_slice(dicts=db_data,
                                                                columns=columns)

            return db_data_filtered
        except:
            raise StorageException('Failed to get data from storage')

    def get_data_by_column(self, collection: str, by: str, value: str, columns: list = [], count: int = 0) -> list:
        '''
        Getting list of documents from the certain collection by requested column: value

        Keyword arguments:
        ---
        collection : str, required
            collection name from the storage
        columns : list, optional (default [])
            list of columns to get values from
        by: str, required
            column name to search by
        value: str, required
            value from the column
        '''
        try:
            db_collection = self.__database[collection]

            if count:
                db_data = list(db_collection.find({by: value}))
            else:
                db_data = list(db_collection.find({by: value}).limit(count))

            db_data_filtered = DataFiltering.dict_list_slice(dicts=db_data,
                                                                columns=columns)

            return db_data_filtered
        except:
            raise StorageException('Failed to get data from storage')

    #---------------------------------------------------------------------------
    # Insertion methods
    #---------------------------------------------------------------------------

    def insert_one_doc(self, collection: str, doc: dict) -> None:
        '''
        Inserts document to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            document to insert to the database
        '''
        try:
            res = self.__database[collection].insert_one(doc)
        except:
            raise StorageException('Failed to insert a doc to collection')

    def insert_many_docs(self, collection: str, docs: list) -> None:
        '''
        Inserts list of documents to the collection

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        docs : list<dict>, required
            listr of documents to insert to the database
        '''
        try:
            res = self.__database[collection].insert_many(docs)
        except:
            raise StorageException('Failed to insert docs to collection!')

    #---------------------------------------------------------------------------
    # Removal methods
    #---------------------------------------------------------------------------

    def remove_one_doc_by_column(self, collection: str, column: str, value: str) -> None:
        '''
        Removes first found document with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        try:
            res = self.__database[collection].delete_one({column: value})
        except:
            raise StorageException('Failed to remove a doc from collection!')

    def remove_one_doc_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes first found document containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        try:
            res = self.__database[collection].delete_one(doc)
        except:
            raise StorageException('Failed to remove a doc from collection!')

    def remove_many_docs_by_column(self, collection: str, column: str, value: str):
        '''
        Removes all documents with specified value from the specified column

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        column : str, required
            column name
        value : str, required
            value from the column
        '''
        try:
            res = self.__database[collection].delete_many({column: value})
        except:
            raise StorageException('Failed to remove a doc from collection!')

    def remove_many_docs_by_dict(self, collection: str, doc: dict) -> None:
        '''
        Removes all documents containg values from specified dictionary

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        doc : dict, required
            "collection_column_name - value" dictionary
        '''
        try:
            res = self.__database[collection].delete_many(doc)
        except:
            raise StorageException('Failed to remove a doc from collection!')

    #---------------------------------------------------------------------------
    # Update methods
    #---------------------------------------------------------------------------

    def update_one_doc(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates first found doc in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for the found doc by id
        '''
        try:
            res = self.__database[collection].update_one({id_column: id}, {'$set': doc}, upsert=True)
        except:
            raise StorageException('Failed to update a doc from collection!')

    def update_many_docs(self, collection: str, id_column: str, id: str, doc: dict) -> None:
        '''
        Updates all documents in certain collection by chosen column by dictionary from args

        Keyword arguments:
        ---
        collection : str, required
            collection name from the database
        id_column : str, required
            name of the id column
        id : str, required
            id value itself
        doc : dict, required
            dictionay with values to update for every found document by id
        '''
        try:
            res = self.__database[collection].update_many({id_column: id}, {'$set': doc}, upsert=True)
        except:
            raise StorageException('Failed to update a doc from collection!')




from typing import Callable, List

from .route import Route


class CommandRoute(Route):
    """
    Route serving incoming commands.

    ..........
    Attributes
    ----------
    command: str, public
        required incoming command to serve the route
    """
    def __init__(self,
                 command: str,
                 callback: Callable,
                 states: List[str],
                 roles: List[str]) -> None:
        """
        Constructor.

        .........
        Arguments
        ---------
        command: str, requited
            trigger command to serve the route
        callback: Callable, required
            function to be called when route is served
        states: List[str]
            states one of which user should have to access serving the route
        roles: List[str]
            roles one of which user should have to access serving the route
        """
        super().__init__(callback, states, roles)

        self.command = command



from ..utils.filemanager import FileManager

from ..storage_managing.storage import Storage

from .exceptions.passworderror import PasswordError
from .exceptions.roleerror import RoleError
from .exceptions.alreadyloggederror import AlreadyLoggedError


class RoleAuth:
    '''
    A class made for simplifiying users' roles management
    '''
    def __init__(self, storage: Storage, roles: dict, users_collection: str = "Users",
                id_column: str = "_id", roles_column: str = "Roles"):
        '''
        Parameters
        ----------
        storage : Storage
            pointer to the inherited class from Storage (i.e MongoDBStorage)
        roles : dict
            a dictionary of roles
            ...
            example:
                {
                    "admin": {
                        "password": "admin-pass"
                    },
                    "user": {
                        "password": ""
                    }
                }
            ...
        users_collection : str, optional
            name of id column from users_collection, by default "Users"
        id_column : str, optional
            name of id column from users_collection, by default "_id"
        roles_column : str, optional
            name of roles column from users_collection, by default "Roles"
        '''
        self.__users_collection = users_collection
        self.__id_column = id_column
        self.__roles_column = roles_column

        self.__storage = storage
        self.__roles = roles

    def __role_check(self, role: str) -> None:
        '''
        Checks if role exists
        Parameters
        ----------
        role : str
            role name to check

        Raises
        ------
        RoleError
            Exception raises when role does not exist
        '''
        if not role in self.__roles:
            raise RoleError("Role is not found!")

    def __password_check(self, role: str, password: str) -> None:
        if self.__roles[role]['password'] != password:
            raise PasswordError("Wrong password!")

    def get_list_of_roles(self) -> list:
        '''
        Gives list of all available roles

        Returns
        -------
        list
            list of roles
        '''
        return list(self.__roles.keys())

    def add_role(self, role: str, password: str = "") -> None:
        '''
        Adds role to the list of roles

        Parameters
        ----------
        role : str
            Role's name
        password : str, optional
            Password for the role, by default ""
        '''
        self.__roles[role] = {}
        self.__roles[role]['password'] = password

    def remove_role(self, role: str, password: str = "") -> None:
        '''
        Removes specified role from the roles list

        Parameters
        ----------
        role : str
            Role's name
        password : str, optional
            Password for the role, by default ""
        '''
        self.__role_check(role)
        self.__password_check(role, password)

        self.__roles.pop(role)

#    def rename_role(self, role: str) -> None:
#        pass

    def get_user_roles(self, user_id: int) -> list:
        '''
        Gives a list of requested user's roles

        Parameters
        ----------
        user_id : int
            user's id from the storage

        Returns
        -------
        list
            list of user's roles

        Raises
        ------
        RoleError
            Raises an exception when user was not found
        '''
        roles = self.__storage.get_data_by_column(collection=self.__users_collection,
                                                    by=self.__id_column,
                                                    value=user_id)
        if not roles:
            raise RoleError("User not found!")

        role = roles[0]

        return role[self.__roles_column]

    def is_loggined_as(self, role: str, user_id: int) -> bool:
        '''
        Checks if user is logged in as requested role

        Parameters
        ----------
        role : str
            Role's name
        user_id : int
            user's id from the storage

        Returns
        -------
        bool
            whether user is logged in or not
        '''
        user_roles = self.get_user_roles(user_id)

        return True if role in user_roles else False

    def login_as(self, role: str, user_id: int, password: str="") -> None:
        '''
        Logs in specified user as a specified role with

        Parameters
        ----------
        role : str
            Role's name
        user_id : int
            User's id from the storage
        password : str
            Roles's password

        Raises
        ------
        AlreadyLoggedError
            Raises an exception if user is already logged in
        '''
        self.__role_check(role)
        self.__password_check(role, password)

        user_roles = self.get_user_roles(user_id)

        if role not in user_roles:
            user_roles.append(role)

            roles_dict = {
                self.__roles_column: user_roles
            }

            self.__storage.update_one_doc(collection=self.__users_collection,
                                        id_column=self.__id_column,
                                        id=user_id,
                                        doc=roles_dict)

        else:
            raise AlreadyLoggedError(f'User has already logged in as a(n) {role}!')

    def logout_as(self, role: str, user_id: int) -> None:
        '''
        Logs the user out

        Parameters
        ----------
        role : str
            Role's name
        user_id : int
            User's id from the storage

        Raises
        ------
        AlreadyLoggedError
            Raises an exception if user was not logged in
        '''
        self.__role_check(role)

        user_roles = self.get_user_roles(user_id)
        if role in user_roles:
            user_roles.remove(role)

            roles_dict = {
                self.__roles_column: user_roles
            }

            self.__storage.update_one_doc(collection=self.__users_collection,
                                        id_column=self.__id_column,
                                        id=user_id,
                                        doc=roles_dict)

        else:
            raise AlreadyLoggedError(f'User has already logged out as a(n) {role}!')



import json
import os
from typing import List, Union


class FileManager:
    """
    Static auxiliary  class to simplify data processing

    ..........
    Attributes
    encoding: str, public
        encoding to open files with

    .......
    Methods
    -------
    clean_file(file_path)
        removes all data from given file
    copy_content(src_path, dest_path)
        if src exists, copies its data into dest
    file_exists(file_path)
        checks if given file exists
    get_dir(file_path)
        returns directory path of given file
    is_file(path)
        checks if given path belongs to file
    list_files(dir_path)
        lists all files in given dir (not recursively)
    make_dir(dir_path)
        creates directory with given path
    make_file(file_path)
        creates file with given path
    move_content(src_path, dest_path)
        if src exists, moves its data into dest
    read_json(file_path)
        reads given json file and returns its content
    remove_file(file_path)
        removes given file
    """

    encoding = "utf-8"

    @staticmethod
    def clean_file(file_path: str) -> None:
        """
        Removes all data from given file. File must exist.

        Arguments
        ---
        file_path: str, required
            path of file to clean
        """
        open(file_path, 'w').close()

    @staticmethod
    def copy_content(src_path: str, dest_path: str) -> None:
        """
        If src exists, copites its data into dest.
        If there is data in dest, it will be replaced with new data.

        Arguments
        ---
        src_path: str, required
            path of source file
        dest_path: str, required
            path of destination file
        """
        if not FileManager.file_exists(src_path):
            return

        with open(src_path, 'r') as src_fd:
            with open(dest_path, 'w') as dest_fd:
                dest_fd.writelines(src_fd.readlines())

    @staticmethod
    def file_exists(file_path: str) -> bool:
        """
        Checks if given file exists.

        Arguments
        ---
        file_path: str, required
            path of file to check
        """
        return os.path.exists(file_path)

    @staticmethod
    def get_dir(file_path: str) -> str:
        """
        Returns directory path of given file. File must exist.

        Arguments
        ---
        file_path: str, required
            path of file to get directory path of
        """
        return os.path.dirname(file_path)

    @staticmethod
    def is_file(path: str) -> bool:
        """
        Checks if given path belongs to file

        Arguments
        ---
        path: str, required
            path to check
        """
        return os.path.isfile(path)

    @staticmethod
    def list_files(dir_path: str) -> List[str]:
        """
        Lists all files in given directory (not recursively). Full path list is returned.

        Arguments
        ---
        dir_path: str, required
            path of directory to list files from
        """
        return [dir_path + '/' + content for content in os.listdir(dir_path)\
            if FileManager.is_file(dir_path + '/' + content)]

    @staticmethod
    def make_dir(dir_path: str) -> None:
        """
        Creates directory with given path. If directory exists nothing happens.

        Arguments
        ---
        dir_path: str, required
            path of directory to make
        """
        os.makedirs(dir_path, exist_ok=True)

    @staticmethod
    def make_file(file_path: str) -> None:
        """
        Creates file with given path. If file already exists nothing happens.

        Arguments
        ---
        file_path: str, required
            path of file to make
        """
        dir_path = FileManager.get_dir(file_path)
        if dir_path:
            FileManager.make_dir(dir_path)
        open(file_path, 'a').close()

    @staticmethod
    def move_content(src_path: str, dest_path: str) -> None:
        """
        If src exists, moves its data into dest.
        If dest contains any data, it will be replaced by new data.

        Arguments
        ---
        src_path: str, required
            path of source file
        dest_path: str, required
            path of destonation file
        """
        FileManager.copy_content(src_path, dest_path)
        FileManager.clean_file(src_path)

    @staticmethod
    def read_json(file_path: str) -> Union[list, dict, None]:
        """
        Reads given json file and returns its content. In case of error returns None.

        Arguments
        ---
        file_path: str, required
            path of json file to read
        """
        if not FileManager.file_exists(file_path):
            return
        with open(file_path, 'r', encoding=FileManager.encoding) as fd:
            try:
                return json.load(fd)
            except Exception as e:
                return

    @staticmethod
    def remove_file(file_path: str) -> None:
        """
        Removes given file. If file does not exist, nothing happens.

        Arguments
        ---
        file_path: str, required
            path of file to remove
        """
        if FileManager.file_exists(file_path):
            os.remove(file_path)




import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="advanced-telegram-bot", # Replace with your own username
    version="0.1.5",
    author="minish144 & usual-one",
    author_email="varlamow.col@yahoo.com",
    description="Python library containing utils for telegram bots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sdallaboratory/advanced-telegram-bot",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['pymongo', 'python-telegram-bot', 'decorator']
)



from .roleerror import RoleError


class AlreadyLoggedError(RoleError):
    '''
    Simple exception class for situations when user had already logged in/out but
    somewhy tried to log in/out once again..
    '''



import copy
from typing import Any

from ..utils.filemanager import FileManager


class LocaleManager:
    """
    Class for locale-dependent content loading and managing

    ...
    Attributes
    ---
    locales: dict
        dictionary with locales ("locale_name": <locale>)

    Methods
    ---
    load_locales(): private
        loads all found locales json files from given folder into locales attribute
    load_locale(): private
        loads locale from given path (must be json file)
    get_format_reply(): public
        gets reply with given name which requires formatting
    get_keyboard(): public
        gets keyboard as list of buttons
    get_keyboard_button(): public
        gets button with given name
    get_other() public
        gets other locale-dependent data with given name
    get_simple_reply(): public
        gets reply with given name which does not require formatting
    """
    def __init__(self, locales_folder: str) -> None:
        """
        Arguments:
        ---
        locales_folder: str, required
            path to folder that contains locales json files
        """
        self.__locales = dict()
        self.__load_locales(locales_folder)

    def __load_locales(self, locales_folder: str) -> None:
        """
        Loads all found locales json files from given folder into locales attribute.
        Locale json file must contain key "code" which value is referred to as
        locale argument in public methods

        Arguments
        ---
        locales_folder: str, required
            path to folder that contains locales json files
        """
        for locale_path in FileManager.list_files(locales_folder):
            locale = self.__load_locale(locale_path)
            self.__locales[locale['code']] = locale

    def __load_locale(self, locale_path: str) -> dict:
        """
        Loads locale from given path (must be json file)

        Arguments
        ---
        locale_path: str, required
            path to locale json file
        """
        return FileManager.read_json(locale_path)

    def get_format_reply(self, name: str, locale: str) -> str:
        """
        Gets reply with given name which requires formatting
        Loaded locales must contain key "replies", "format" inside of "replies" value
        and name inside of "format" value

        Arguments
        ---
        name: str, required
            name of reply (which is key of reply inside "format" value)
        locale: str, required
            locale code
        """
        return self.__locales[locale]['replies']['format'][name]

    def get_keyboard(self, name: str, locale: str) -> list:
        """
        Gets keyboard as list of buttons.
        Locale file must contain 'keyboards':'arrangements':'<name>' structure.

        .........
        Arguments
        ---------
        name: str, required
            name of the keyboard
        locale: str, required
            locale code
        """
        buttons = copy.deepcopy(self.__locales[locale]['keyboards']['arrangements'][name])
        for i in range(len(buttons)):
            for j in range(len(buttons[i])):
                buttons[i][j] = self.get_keyboard_button(buttons[i][j], locale)
        return buttons

    def get_keyboard_button(self, button: str, locale: str) -> str:
        """
        Gets button with given name.
        Locale file must contain 'keyboards':'buttons':'<button>' structure.

        .........
        Arguments
        ---------
        button: str, required
            button name
        locale: str, required
            locale code
        """
        return self.__locales[locale]['keyboards']['buttons'][button]

    def get_other(self, name: str, locale: str) -> Any:
        """
        Gets other locale-dependent data with given name
        Loaded locales must contain key "other" and name inside of "replies" value

        Arguments
        ---
        name: str, required
            name of locale-dependent data (which is key inside "other" value)
        locale: str, required
            locale code
        """
        return self.__locales[locale]['other'][name]

    def get_simple_reply(self, name: str, locale: str) -> str:
        """
        Gets reply with given name which does not require formatting
        Loaded locales must contain key "replies", "format" inside of "replies" value
        and name inside of "simple" value

        Arguments
        ---
        name: str, required
            name of reply (which is key of reply inside "simple" value)
        locale: str, required
            locale code
        """
        return self.__locales[locale]['replies']['simple'][name]

